#include "changer.h"

#include <windows.h>
#include <iostream>
#include <iterator>
#include <algorithm>
#include <sstream>
#include <memory>

#include "display.h"
#include "window.h"
#include "config.h"

// missing in current mingw build
#ifndef MOD_NOREPEAT
#define MOD_NOREPEAT 0x4000
#endif

namespace {

// std::to_string not available in current mingw build
std::string toString(int i) {
	std::stringstream out{};
	out << i;
	return out.str();
}

void CALLBACK hideWindow(HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime) {
	std::cout << "hide fired" << std::endl;
	Window::getWindow(hwnd).hide();
	KillTimer(hwnd, 0);
}

void showText(Window& window, std::string const& text) {
	window.setText(text);
	SetTimer(window.getHandle(), 0, 5000, static_cast<TIMERPROC>(&hideWindow));
	window.show();
}

DWORD getNewRefreshRate(Display& display, Config const& config) {
	auto rate = std::find(config.frequencies.begin(), config.frequencies.end(), display.getRefreshRate());
	if(rate == config.frequencies.end()) {
		rate = config.frequencies.begin();
	} else {
		rate++;
		if(rate == config.frequencies.end()) {
			rate = config.frequencies.begin();
		}
	}
	return *rate;
}

void changeRefreshRate(Window& window, Display& display, Config const& config) {
	DWORD newRefreshRate = getNewRefreshRate(display, config);
	std::cout << "updating refresh rate: " << newRefreshRate << std::endl;
	if(display.setRefreshRate(newRefreshRate)) {
		showText(window, toString(display.getRefreshRate()) + " Hz");
	} else {
		std::cout << "error changing refresh rate, " << newRefreshRate << " Hz not supported?" << std::endl;
		showText(window, "Error");
	}
}

MSG runEventLoop(Window& window, Display& display, Config const& config) {
	// shows the startup spash screen :)
	showText(window, "FQCHG");
	
	MSG msg{};
	while(GetMessage(&msg, NULL, 0, 0) > 0) {
		switch(msg.message) {
		case WM_HOTKEY:
			changeRefreshRate(window, display, config);
			break;
		case WM_TIMER:
			// I preferred to use this case over the function, but
			// somehow this event is fired not to the correct time (especially for the first execution)
			//std::cout << "hiding window\n";
			//window.hide();
			break;
		}
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return msg;
}

}

namespace changer {

int run(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
	RegisterHotKey(NULL, 1, MOD_ALT | MOD_NOREPEAT, 0x42);	// alt+b, 0x42=ASCII Code 'B'
	//std::cout << lpCmdLine << '\n';

	auto displays = Display::getDisplays();
	std::cout << "Available devices: \n";
	std::copy(displays.begin(), displays.end(), std::ostream_iterator<Display>{std::cout, "\n"});

	auto it = std::find_if(displays.begin(), displays.end(), [](Display const& d) { return d.isPrimary(); });
	if(it != displays.end()) {
		auto refreshRates = (*it).getSupportedRefreshRates();
		std::cout << "Using:\n" << *it << "\nSupported Refresh Rates: ";
		std::copy(refreshRates.begin(), refreshRates.end(), std::ostream_iterator<DWORD>{std::cout, " Hz "});
		std::cout << std::endl;

		Window window{hInstance, nCmdShow, 240, 120};
		return runEventLoop(window, *it, Config{"config.ini"}).wParam;
	}
	std::cout << "No primary display found";
	return 1;
}

}
