#include "display.h"

#include <windows.h>
#include <list>
#include <set>
#include <memory>
#include <ostream>
#include <stdexcept>
//#include <iostream>

// Define DISPLAY_DEVICE_ACTIVE as it is not defined in MinGW
#ifndef DISPLAY_DEVICE_ACTIVE
#define DISPLAY_DEVICE_ACTIVE    0x00000001
#endif

std::list<Display> Display::getDisplays() {
	std::list<Display> displays;
	//int index{};
	DISPLAY_DEVICE displayDevice{};
	displayDevice.cb = sizeof(DISPLAY_DEVICE);
	
	//while(EnumDisplayDevices(NULL, index++, &displayDevice, 0)) {
	for(int i{}; EnumDisplayDevices(NULL, i, &displayDevice, 0); ++i) {
		displays.push_back(Display{displayDevice});
		displayDevice = DISPLAY_DEVICE{};
		displayDevice.cb = sizeof(DISPLAY_DEVICE);
	}
	return displays;
}

Display::Display(DISPLAY_DEVICE dp):displayDevice(dp) {
}

bool Display::isPrimary() const {
	return displayDevice.StateFlags & DISPLAY_DEVICE_PRIMARY_DEVICE;
}

bool Display::isActive() const {
	return displayDevice.StateFlags & DISPLAY_DEVICE_ACTIVE;
}

DWORD Display::getRefreshRate() const {
	return getDisplaySettings().dmDisplayFrequency;
}

bool Display::setRefreshRate(DWORD refreshRate) {
	auto deviceMode = getDisplaySettings();
	deviceMode.dmDisplayFrequency = refreshRate;
	return ChangeDisplaySettings(&deviceMode, 0) == DISP_CHANGE_SUCCESSFUL;
}

std::set<DWORD> Display::getSupportedRefreshRates() const {
	std::set<DWORD> refreshRates{};
	DEVMODE deviceMode{};
	deviceMode.dmSize = sizeof(DEVMODE);
	//int index{};

	DEVMODE activeDeviceMode = getDisplaySettings();
	//while(EnumDisplaySettings(displayDevice.DeviceName, index++, &deviceMode)) {
	for(int i{}; EnumDisplaySettings(displayDevice.DeviceName, i, &deviceMode); ++i) {
		// We skip different resolutions as there is a slight possibility that one
		// resolution supports other refresh rates than the current.
		//std::cout << deviceMode.dmPelsWidth << 'x' << deviceMode.dmPelsHeight << '=';
		//std::cout << activeDeviceMode.dmPelsWidth << 'x' << activeDeviceMode.dmPelsHeight << '\n';
		if(activeDeviceMode.dmPelsWidth != deviceMode.dmPelsWidth
				|| activeDeviceMode.dmPelsHeight != deviceMode.dmPelsHeight) {
			continue;
		}
		refreshRates.insert(deviceMode.dmDisplayFrequency);
	}
	return refreshRates;
}

DEVMODE Display::getDisplaySettings() const {
	DEVMODE deviceMode{};
	deviceMode.dmSize = sizeof(DEVMODE);
	EnumDisplaySettings(displayDevice.DeviceName, ENUM_CURRENT_SETTINGS, &deviceMode);
	return deviceMode;
}

std::ostream& Display::write(std::ostream& out) const {
	return out << displayDevice.DeviceName << " - " 
		<< displayDevice.DeviceString << " - " << getRefreshRate() << " Hz"
		<< " - " << (isActive() ? "active" : "inactive")
		<< (isPrimary() ? " - primary" : "");
}

std::ostream& operator<<(std::ostream& lhv, Display const& rhv) {
	return rhv.write(lhv);
}
